@@ -190,7 +190,8 @@ def try_import(mod: str, only_modules=False) -> List[str]:
         completions.extend(m_all)
 
     if m_is_init:
-        completions.extend(module_list(os.path.dirname(m.__file__)))
+        file_ = m.__file__
+        completions.extend(module_list(os.path.dirname(file_)))
     completions_set = {c for c in completions if isinstance(c, str)}
     completions_set.discard('__init__')
     return list(completions_set)
